from typing import Optional

from clowmdb.models import Bucket, BucketPermission
from sqlalchemy import func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession


class CRUDBucket:
    @staticmethod
    async def check_bucket_exist(db: AsyncSession, bucket_name: str) -> bool:
        """
        Check if the given bucket exists.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        bucket_name : str
            Name of a bucket.

        Returns
        -------
        check : bool
            Flag if the check was successful.
        """
        stmt = select(Bucket).where(Bucket.name == bucket_name)
        bucket = await db.scalar(stmt)
        return bucket is not None

    @staticmethod
    async def check_access(db: AsyncSession, bucket_name: str, uid: str, key: Optional[str] = None) -> bool:
        """
        Check if the given user has access to the bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        bucket_name : str
            Name of a bucket.
        uid : str
            UID of a user.
        key : str | None, default None
            Additional check if the user has access to the key in the bucket.

        Returns
        -------
        check : bool
            Flag if the check was successful.
        """
        stmt = select(Bucket).where(Bucket.name == bucket_name).where(Bucket.owner_id == uid)
        bucket = await db.scalar(stmt)
        # If the user is the owner of the bucket -> user has access
        if bucket is not None:
            return True

        stmt = (
            select(BucketPermission)
            .where(BucketPermission.bucket_name == bucket_name)  # check bucket name
            .where(BucketPermission.user_id == uid)  # check grantee of permission
            .where(BucketPermission.permissions == BucketPermission.Permission.READWRITE)  # check READWRITE Permission
            .where(  # check 'form' timestamp is no violated
                or_(
                    func.UNIX_TIMESTAMP() >= BucketPermission.from_,
                    BucketPermission.from_ == None,  # noqa:E711
                )
            )
            .where(  # check 'to' timestamp is no violated
                or_(
                    func.UNIX_TIMESTAMP() <= BucketPermission.to,
                    BucketPermission.to == None,  # noqa:E711
                )
            )
        )

        permission: Optional[BucketPermission] = await db.scalar(stmt)
        # If the user has no active READWRITE Permission for the bucket -> user has no access
        if permission is None:
            return False
        # If the permission has no restriction on the file prefix -> user has access
        if permission.file_prefix is None:
            return True
        # If the permission has restriction on the file prefix but no key is given -> user has no access
        if key is None:
            return False

        # If the permission has restriction on the file prefix and key is given -> check if key starts with file prefix
        return key.startswith(permission.file_prefix)
