from typing import Iterable, List, Optional, Union
from uuid import UUID

from clowmdb.models import WorkflowMode, workflow_mode_association_table
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession

from app.schemas.workflow_mode import WorkflowModeIn


class CRUDWorkflowMode:
    @staticmethod
    async def list_modes(
        db: AsyncSession,
        workflow_version: str,
    ) -> List[WorkflowMode]:
        """
        List all workflow modes of a specific workflow version.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        workflow_version : str
            The version id for which the modes should be loaded.

        Returns
        -------
        modes : List[clowmdb.models.WorkflowMode]
            List of workflow modes.
        """
        stmt = (
            select(WorkflowMode)
            .join(workflow_mode_association_table)
            .where(workflow_mode_association_table.columns.workflow_version_commit_hash == workflow_version)
        )
        return list((await db.scalars(stmt)).all())

    @staticmethod
    async def get(
        db: AsyncSession, mode_id: Union[bytes, UUID], workflow_version: Optional[str] = None
    ) -> Optional[WorkflowMode]:
        """
        Get a specific workflow mode.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        mode_id : UUID | bytes
            ID of a workflow mode.
        workflow_version : str | None, default None
            Optional workflow version the workflow mode has to be connected to.

        Returns
        -------
        workflows : clowmdb.models.WorkflowMode | None
            Requested workflow mode if it exists, None otherwise
        """
        mid = mode_id.bytes if isinstance(mode_id, UUID) else mode_id
        stmt = select(WorkflowMode).where(WorkflowMode._mode_id == mid)
        if workflow_version is not None:
            stmt = stmt.join(workflow_mode_association_table).where(
                workflow_mode_association_table.columns.workflow_version_commit_hash == workflow_version
            )
        return await db.scalar(stmt)

    @staticmethod
    async def create(db: AsyncSession, modes: List[WorkflowModeIn]) -> List[WorkflowMode]:
        """
        Create multiple workflow modes at once.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        modes : List[app.schemas.workflow_mode.WorkflowModeIn]
            Schemas of the workflow modes to be created.

        Returns
        -------
        modes : List[clowmdb.models.WorkflowMode]
            Newly created workflow modes
        """
        modes_db = []
        for mode in modes:
            mode_db = WorkflowMode(name=mode.name, entrypoint=mode.entrypoint, schema_path=mode.schema_path)
            db.add(mode_db)
            modes_db.append(mode_db)
        await db.commit()
        return modes_db

    @staticmethod
    async def delete(db: AsyncSession, modes: Iterable[UUID]) -> None:
        """
        Delete multiple workflows at once.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        modes : List[uuid.UUID]
            ID of workflow modes to delete
        """
        stmt = delete(WorkflowMode).where(WorkflowMode._mode_id.in_([uuid.bytes for uuid in modes]))
        await db.execute(stmt)
        await db.commit()
