from functools import cached_property
from typing import Dict, Optional
from urllib.parse import quote, urlparse

from fastapi import status
from httpx import USE_CLIENT_DEFAULT, AsyncClient, BasicAuth
from pydantic import AnyHttpUrl

from .abstract_repository import GitRepository


class GitHubRepository(GitRepository):
    """
    Implementation for GitHub Repository
    """

    @property
    def token(self) -> Optional[str]:
        return self._token

    @property
    def provider(self) -> str:
        return "github"

    @property
    def repo_url(self) -> str:
        return self.url

    @property
    def git_commit_hash(self) -> str:
        return self.commit

    @cached_property
    def request_auth(self) -> Optional[BasicAuth]:
        if self._token is not None:
            return BasicAuth(username=self.user, password=self._token)
        return None

    @cached_property
    def request_headers(self) -> Dict[str, str]:
        return {"Accept": "application/vnd.github.object+json", "X-GitHub-Api-Version": "2022-11-28"}

    def __init__(self, url: str, git_commit_hash: str, token: Optional[str] = None, user: Optional[str] = None):
        parse_result = urlparse(url)
        path_parts = parse_result.path[1:].split("/")
        self.url = url
        self.account = path_parts[0]
        self.repository = path_parts[1]
        self.commit = git_commit_hash
        self._token = token
        self.user = user if user else self.account

    def check_file_url(self, filepath: str) -> AnyHttpUrl:
        return AnyHttpUrl.build(
            scheme="https",
            host="api.github.com",
            path="/".join(["repos", self.account, self.repository, "contents", quote(filepath)]),
            query=f"ref={quote(self.commit)}",
        )

    async def download_file_url(self, filepath: str, client: AsyncClient) -> AnyHttpUrl:
        # If the repo is public, build download URL directly
        if self._token is None:
            return AnyHttpUrl.build(
                scheme="https",
                host="raw.githubusercontent.com",
                path="/".join([self.account, self.repository, self.commit, filepath]),
            )
        # If the repo is private, request a download URL with a token from the GitHub API
        response = await client.get(
            str(self.check_file_url(filepath)),
            auth=USE_CLIENT_DEFAULT if self.request_auth is None else self.request_auth,
            headers=self.request_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        return AnyHttpUrl(response.json()["download_url"])

    def __repr__(self) -> str:
        url = AnyHttpUrl.build(scheme="https", host="github.com", path="/".join([self.account, self.repository]))
        return f"GitHub(repo={url} commit={self.commit})"
