from os import environ
from typing import Optional
from uuid import UUID

import dotenv
from httpx import AsyncClient

from app.core.config import settings

dotenv.load_dotenv()
base_env = {key: val for key, val in environ.items() if key.startswith("NXF_")}


class SlurmClient:
    def __init__(self, client: AsyncClient, version: str = "v0.0.38"):
        """
        Initialize the client to communicate with a Slurm cluster.

        Parameters
        ----------
        client : httpx.AsyncClient
            Async HTTP client with an open connection.
        version : str, default "v0.0.38"
            Version of the Slurm Rest API
        """
        self._client = client
        self._headers = {"X-SLURM-USER-TOKEN": settings.SLURM_TOKEN, "X-SLURM-USER-NAME": settings.SLURM_USER}
        self.version = version

    async def submit_job(self, nextflow_script: str, execution_id: UUID, scm_file_name: Optional[str] = None) -> int:
        """
        Submit a job to the slurm cluster.

        Parameters
        ----------
        nextflow_script : str
            Script to execute on the slurm cluster.
        execution_id : uuid.UUID
            ID of the workflow execution.

        Returns
        -------
        slurm_job_id : int
            Slurm job ID of submitted job.
        """
        env = base_env.copy()
        env["TOWER_WORKSPACE_ID"] = execution_id.hex[:16]
        # Reference a SCM file for a private repository
        if scm_file_name is not None:
            env["NXF_SCM_FILE"] = f"{settings.PARAMS_BUCKET_MOUNT_PATH}/{scm_file_name}"

        body = {
            "script": nextflow_script,
            "job": {
                "name": str(execution_id),
                "current_working_directory": settings.SLURM_WORKING_DIRECTORY,
                "requeue": False,
                "environment": env,
            },
        }
        response = await self._client.post(
            f"{settings.SLURM_ENDPOINT}slurm/{self.version}/job/submit", headers=self._headers, json=body
        )
        return int(response.json()["job_id"])

    async def cancel_job(self, job_id: int) -> None:
        """
        Cancel a Slurm job on the cluster.

        Parameters
        ----------
        job_id : int
            ID of the job to cancel.
        """
        await self._client.delete(f"{settings.SLURM_ENDPOINT}/slurm/{self.version}/job/{job_id}", headers=self._headers)
