from io import BytesIO
from uuid import UUID, uuid4

import pytest
from clowmdb.models import Bucket, Workflow, WorkflowExecution, WorkflowMode, WorkflowVersion
from fastapi import status
from httpx import AsyncClient
from sqlalchemy import update
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.config import settings
from app.schemas.workflow import WorkflowOut
from app.schemas.workflow_execution import DevWorkflowExecutionIn, WorkflowExecutionIn
from app.scm import SCM
from app.tests.mocks.mock_s3_resource import MockS3ServiceResource
from app.tests.utils.bucket import add_permission_for_bucket
from app.tests.utils.user import UserWithAuthHeader
from app.tests.utils.utils import random_hex_string, random_lower_string


class _TestWorkflowExecutionRoutes:
    base_path: str = "/workflow_executions"


class TestWorkflowExecutionRoutesCreate(_TestWorkflowExecutionRoutes):
    @pytest.mark.asyncio
    async def test_start_github_workflow_execution(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        execution_in = WorkflowExecutionIn(workflow_version_id=random_workflow_version.git_commit_hash, parameters={})
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["workflow_version_id"] == execution_in.workflow_version_id
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING
        assert execution_response["workflow_version_id"] == random_workflow_version.git_commit_hash

        assert (
            f"params-{UUID(hex=execution_response['execution_id']).hex }.json"
            in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        )

    @pytest.mark.asyncio
    async def test_start_private_github_workflow_execution(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
        random_private_workflow: WorkflowOut,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution from a private GitHub repository.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        random_private_workflow : app.schemas.workflow.WorkflowOut
            Random private workflow for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """

        execution_in = WorkflowExecutionIn(workflow_version_id=random_workflow_version.git_commit_hash, parameters={})
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["workflow_version_id"] == execution_in.workflow_version_id
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING
        assert execution_response["workflow_version_id"] == random_workflow_version.git_commit_hash

        assert (
            f"params-{UUID(hex=execution_response['execution_id']).hex }.json"
            in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        )

    @pytest.mark.asyncio
    async def test_start_gitlab_workflow_execution(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution.

        Parameters
        ----------
        db :
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        stmt = (
            update(Workflow)
            .where(Workflow._workflow_id == random_workflow_version.workflow_id.bytes)
            .values(repository_url="https://gitlab.com/example/example")
        )
        await db.execute(stmt)
        await db.commit()
        execution_in = WorkflowExecutionIn(workflow_version_id=random_workflow_version.git_commit_hash, parameters={})
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["workflow_version_id"] == execution_in.workflow_version_id
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING
        assert execution_response["workflow_version_id"] == random_workflow_version.git_commit_hash

        assert (
            f"params-{UUID(hex=execution_response['execution_id']).hex }.json"
            in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        )

    @pytest.mark.asyncio
    async def test_start_too_many_workflow_executions(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting too many workflow executions.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        active_execution_counter = 0
        while active_execution_counter < settings.ACTIVE_WORKFLOW_EXECUTION_LIMIT:
            execution = WorkflowExecution(
                user_id=random_user.user.uid,
                workflow_version_id=random_workflow_version.git_commit_hash,
                slurm_job_id=1,
            )
            db.add(execution)
            active_execution_counter += 1
        await db.commit()
        execution_in = WorkflowExecutionIn(workflow_version_id=random_workflow_version.git_commit_hash, parameters={})
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_unknown_workflow_version(
        self, client: AsyncClient, random_user: UserWithAuthHeader
    ) -> None:
        """
        Test for starting a workflow execution with an unknown workflow version.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_hex_string(),
            parameters={},
        )
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_deprecated_workflow_version(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting a workflow execution with an unknown workflow version.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        await db.execute(
            update(WorkflowVersion)
            .where(WorkflowVersion.git_commit_hash == random_workflow_version.git_commit_hash)
            .values(status=WorkflowVersion.Status.DEPRECATED.name)
        )
        await db.commit()
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={},
        )
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_non_existing_bucket(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting a workflow execution with a non-existing bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={"dir": "s3://" + random_lower_string()},
        )
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_existing_bucket(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_bucket: Bucket,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting a workflow execution with an existing bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={"dir": f"s3://{random_bucket.name}"},
        )
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_201_CREATED

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_bucket_permission(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_second_user: UserWithAuthHeader,
        random_bucket: Bucket,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting a workflow execution with a permission for the reference bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user for testing.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        await add_permission_for_bucket(db, random_bucket.name, random_second_user.user.uid)
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={"dir": f"s3://{random_bucket.name}"},
        )
        response = await client.post(
            self.base_path, headers=random_second_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_201_CREATED

    @pytest.mark.asyncio
    async def test_start_workflow_execution_without_bucket_permission(
        self,
        client: AsyncClient,
        random_second_user: UserWithAuthHeader,
        random_bucket: Bucket,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting a workflow execution without a permission for the reference bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user for testing.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={"dir": f"s3://{random_bucket.name}"},
        )
        response = await client.post(
            self.base_path, headers=random_second_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_good_report_bucket(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_bucket: Bucket,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting a workflow execution where the report output bucket is an existing bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={},
            report_output_bucket=random_bucket.name,
        )
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_201_CREATED

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_bad_report_bucket(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
    ) -> None:
        """
        Test for starting a workflow execution where the report output bucket is a non-existing bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={},
            report_output_bucket=random_lower_string(),
        )
        response = await client.post(self.base_path, headers=random_user.auth_headers, json=execution_in.model_dump())
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_workflow_mode(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_version: WorkflowVersion,
        random_workflow_mode: WorkflowMode,
    ) -> None:
        """
        Test for starting a workflow execution where the report output bucket is a non-existing bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash,
            parameters={},
            mode=random_workflow_mode.mode_id,
        ).model_dump_json()
        response = await client.post(self.base_path, headers=random_user.auth_headers, data=execution_in)
        assert response.status_code == status.HTTP_201_CREATED

    @pytest.mark.asyncio
    async def test_start_workflow_execution_with_non_existing_workflow_mode(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow_version: WorkflowVersion
    ) -> None:
        """
        Test for starting a workflow execution where the report output bucket is a non-existing bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        """
        execution_in = WorkflowExecutionIn(
            workflow_version_id=random_workflow_version.git_commit_hash, parameters={}, mode=uuid4()
        ).model_dump_json()
        response = await client.post(self.base_path, headers=random_user.auth_headers, data=execution_in)
        assert response.status_code == status.HTTP_404_NOT_FOUND


class TestDevWorkflowExecutionRoutesCreate(_TestWorkflowExecutionRoutes):
    @pytest.mark.asyncio
    async def test_start_dev_workflow_execution_from_github(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution with an arbitrary GitHub repository.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        execution_in = DevWorkflowExecutionIn(
            git_commit_hash=random_hex_string(), repository_url="https://github.com/example-user/example", parameters={}
        )
        response = await client.post(
            f"{self.base_path}/arbitrary", headers=random_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING

        assert (
            f"params-{UUID(hex=execution_response['execution_id']).hex }.json"
            in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        )

    @pytest.mark.asyncio
    async def test_start_dev_workflow_execution_from_gitlab(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution with an arbitrary Gitlab repository.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        execution_in = DevWorkflowExecutionIn(
            git_commit_hash=random_hex_string(), repository_url="https://gitlab.com/example-user/example", parameters={}
        )
        response = await client.post(
            f"{self.base_path}/arbitrary", headers=random_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING

        assert (
            f"params-{UUID(hex=execution_response['execution_id']).hex }.json"
            in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        )

    @pytest.mark.asyncio
    async def test_start_dev_workflow_execution_from_private_gitlab(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution with an arbitrary private Gitlab repository.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        token = random_lower_string(15)
        execution_in = DevWorkflowExecutionIn(
            git_commit_hash=random_hex_string(),
            repository_url="https://gitlab.com/example-user/example",
            parameters={},
            token=token,
        )
        response = await client.post(
            f"{self.base_path}/arbitrary", headers=random_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING

        execution_id = UUID(hex=execution_response["execution_id"])

        # Check if params file is created
        params_file_name = f"params-{execution_id.hex }.json"
        assert params_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()

        # Check if SCM file is created
        scm_file_name = f"{execution_id.hex}.scm"
        assert scm_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        scm_obj = mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(scm_file_name)
        with BytesIO() as f:
            scm_obj.download_fileobj(f)
            f.seek(0)
            scm = SCM.deserialize(f)

        # Check content of SCM file
        assert len(scm.providers) == 1
        provider = scm.providers[0]
        assert provider.password == token
        assert provider.name == f"repo{execution_id.hex}"
        assert provider.platform == "gitlab"
        assert provider.server == "https://gitlab.com"
        # Clean up after test
        mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(params_file_name).delete()
        scm_obj.delete()

    @pytest.mark.asyncio
    async def test_start_dev_workflow_execution_from_private_github_without_username(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution with an arbitrary private GitHub repository without a provided username.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        token = random_lower_string(15)
        execution_in = DevWorkflowExecutionIn(
            git_commit_hash=random_hex_string(),
            repository_url="https://github.com/example-user/example",
            parameters={},
            token=token,
        )
        response = await client.post(
            f"{self.base_path}/arbitrary", headers=random_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING

        execution_id = UUID(hex=execution_response["execution_id"])

        # Check if params file is created
        params_file_name = f"params-{execution_id.hex }.json"
        assert params_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()

        # Check if SCM file is created
        scm_file_name = f"{execution_id.hex}.scm"
        assert scm_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        scm_obj = mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(scm_file_name)
        with BytesIO() as f:
            scm_obj.download_fileobj(f)
            f.seek(0)
            scm = SCM.deserialize(f)

        # Check content of SCM file
        assert len(scm.providers) == 1
        provider = scm.providers[0]
        assert provider.password == token
        assert provider.name == "github"
        assert provider.user == "example-user"
        # Clean up after test
        mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(params_file_name).delete()
        scm_obj.delete()

    @pytest.mark.asyncio
    async def test_start_dev_workflow_execution_from_private_github_with_username(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution with an arbitrary private GitHub repository with a provided username.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        token = random_lower_string(15)
        execution_in = DevWorkflowExecutionIn(
            git_commit_hash=random_hex_string(),
            repository_url="https://github.com/example-user/example",
            parameters={},
            token=token,
            username="bilbobaggins",
        )
        response = await client.post(
            f"{self.base_path}/arbitrary", headers=random_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_201_CREATED
        execution_response = response.json()
        assert execution_response["user_id"] == random_user.user.uid
        assert execution_response["status"] == WorkflowExecution.WorkflowExecutionStatus.PENDING

        execution_id = UUID(hex=execution_response["execution_id"])

        # Check if params file is created
        params_file_name = f"params-{execution_id.hex }.json"
        assert params_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()

        # Check if SCM file is created
        scm_file_name = f"{execution_id.hex}.scm"
        assert scm_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        scm_obj = mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(scm_file_name)
        with BytesIO() as f:
            scm_obj.download_fileobj(f)
            f.seek(0)
            scm = SCM.deserialize(f)

        # Check content of SCM file
        assert len(scm.providers) == 1
        provider = scm.providers[0]
        assert provider.password == token
        assert provider.name == "github"
        assert provider.user == "bilbobaggins"
        # Clean up after test
        mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(params_file_name).delete()
        scm_obj.delete()

    @pytest.mark.asyncio
    async def test_start_dev_workflow_execution_with_unknown_repository(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for starting a workflow execution with an unknown git repository.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        execution_in = DevWorkflowExecutionIn(
            git_commit_hash=random_hex_string(), repository_url="https://bitbucket.com/example/example", parameters={}
        )
        response = await client.post(
            f"{self.base_path}/arbitrary", headers=random_user.auth_headers, json=execution_in.model_dump()
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST


class TestWorkflowExecutionRoutesGet(_TestWorkflowExecutionRoutes):
    @pytest.mark.asyncio
    async def test_get_workflow_execution(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for getting a workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_execution : clowmdb.models.WorkflowExecution
            Random workflow execution for testing.
        """
        response = await client.get(
            "/".join([self.base_path, str(random_workflow_execution.execution_id)]), headers=random_user.auth_headers
        )
        assert response.status_code == status.HTTP_200_OK

        execution = response.json()
        assert execution["execution_id"] == str(random_workflow_execution.execution_id)

    @pytest.mark.asyncio
    async def test_get_workflow_execution_params(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for getting the parameters of a workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_execution : clowmdb.models.WorkflowExecution
            Random workflow execution for testing.
        """
        response = await client.get(
            "/".join([self.base_path, str(random_workflow_execution.execution_id), "params"]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_200_OK

        execution = response.json()
        assert len(execution) == 0

    @pytest.mark.asyncio
    async def test_get_non_existing_workflow_execution(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for getting a non-existing workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        response = await client.get("/".join([self.base_path, str(uuid4())]), headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_404_NOT_FOUND


class TestWorkflowExecutionRoutesList(_TestWorkflowExecutionRoutes):
    @pytest.mark.asyncio
    async def test_list_workflow_executions(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for listing all workflow executions.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_execution : clowmdb.models.WorkflowExecution
            Random workflow execution for testing.
        """
        response = await client.get(self.base_path, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_200_OK

        executions = response.json()
        assert len(executions) > 0
        assert (
            sum(
                1
                for execution in executions
                if execution["execution_id"] == str(random_workflow_execution.execution_id)
            )
            == 1
        )


class TestWorkflowExecutionRoutesDelete(_TestWorkflowExecutionRoutes):
    @pytest.mark.asyncio
    async def test_delete_unfinished_workflow_execution(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for deleting an unfinished workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_execution : clowmdb.models.WorkflowExecution
            Random workflow execution for testing.
        """
        response = await client.delete(
            "/".join([self.base_path, str(random_workflow_execution.execution_id)]), headers=random_user.auth_headers
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_delete_finished_workflow_execution(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for deleting a workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_execution : clowmdb.models.WorkflowExecution
            Random workflow execution for testing.
        """
        await db.execute(
            update(WorkflowExecution)
            .where(WorkflowExecution._execution_id == random_workflow_execution.execution_id.bytes)
            .values(status=WorkflowExecution.WorkflowExecutionStatus.SUCCESS.name)
        )
        await db.commit()
        response = await client.delete(
            "/".join([self.base_path, str(random_workflow_execution.execution_id)]), headers=random_user.auth_headers
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT


class TestWorkflowExecutionRoutesCancel(_TestWorkflowExecutionRoutes):
    @pytest.mark.asyncio
    async def test_cancel_workflow_execution(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for canceling a workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_execution : clowmdb.models.WorkflowExecution
            Random workflow execution for testing.
        """
        response = await client.post(
            "/".join([self.base_path, str(random_workflow_execution.execution_id), "cancel"]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.asyncio
    async def test_cancel_finished_workflow_execution(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for canceling a finished workflow execution.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow_execution : clowmdb.models.WorkflowExecution
            Random workflow execution for testing.
        """
        await db.execute(
            update(WorkflowExecution)
            .where(WorkflowExecution._execution_id == random_workflow_execution.execution_id.bytes)
            .values(status=WorkflowExecution.WorkflowExecutionStatus.SUCCESS.name)
        )
        await db.commit()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow_execution.execution_id), "cancel"]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST
