from io import BytesIO

import pytest
from clowmdb.models import WorkflowVersion
from fastapi import status
from httpx import AsyncClient
from mocks.mock_s3_resource import MockS3ServiceResource
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.endpoints.workflow_version import DocumentationEnum
from app.core.config import settings
from app.schemas.workflow import WorkflowOut
from app.schemas.workflow_version import WorkflowVersionStatus
from app.tests.utils.user import UserWithAuthHeader
from app.tests.utils.utils import random_hex_string


class _TestWorkflowVersionRoutes:
    base_path: str = "/workflows"


class TestWorkflowVersionRoutesGet(_TestWorkflowVersionRoutes):
    @pytest.mark.asyncio
    async def test_get_workflow_version(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for getting a workflow version by its commit hash.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.get(
            "/".join(
                [
                    self.base_path,
                    str(random_workflow.workflow_id),
                    "versions",
                    random_workflow.versions[0].git_commit_hash,
                ]
            ),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        version = response.json()
        assert version["workflow_id"] == str(random_workflow.workflow_id)
        assert version["git_commit_hash"] == random_workflow.versions[0].git_commit_hash

    @pytest.mark.asyncio
    async def test_get_non_existing_version(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for getting a non-existing workflow version.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.get(
            "/".join([self.base_path, str(random_workflow.workflow_id), "versions", random_hex_string()]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND


class TestWorkflowVersionRoutesList(_TestWorkflowVersionRoutes):
    @pytest.mark.asyncio
    async def test_get_workflow_version(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for listing all workflow version from a given workflow.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.get(
            "/".join([self.base_path, str(random_workflow.workflow_id), "versions"]),
            params={"version_status": WorkflowVersion.Status.CREATED.name},
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        versions = response.json()
        assert len(versions) == 1
        version = versions[0]
        assert version["workflow_id"] == str(random_workflow.workflow_id)
        assert version["git_commit_hash"] == random_workflow.versions[0].git_commit_hash


class TestWorkflowVersionRoutesUpdate(_TestWorkflowVersionRoutes):
    @pytest.mark.asyncio
    async def test_update_workflow_version_status(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test updating a workflow version status.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.patch(
            "/".join(
                [
                    self.base_path,
                    str(random_workflow.workflow_id),
                    "versions",
                    random_workflow.versions[0].git_commit_hash,
                    "status",
                ]
            ),
            headers=random_user.auth_headers,
            json=WorkflowVersionStatus(status=WorkflowVersion.Status.PUBLISHED).model_dump(),
        )

        assert response.status_code == status.HTTP_200_OK
        version = response.json()
        assert version["git_commit_hash"] == random_workflow.versions[0].git_commit_hash
        assert version["status"] == WorkflowVersion.Status.PUBLISHED

    @pytest.mark.asyncio
    async def test_deprecate_workflow_version(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test deprecate a workflow version.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.post(
            "/".join(
                [
                    self.base_path,
                    str(random_workflow.workflow_id),
                    "versions",
                    random_workflow.versions[0].git_commit_hash,
                    "deprecate",
                ]
            ),
            headers=random_user.auth_headers,
        )

        assert response.status_code == status.HTTP_200_OK
        version = response.json()
        assert version["git_commit_hash"] == random_workflow.versions[0].git_commit_hash
        assert version["status"] == WorkflowVersion.Status.DEPRECATED

    @pytest.mark.asyncio
    async def test_update_non_existing_workflow_version_status(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test updating a non-existing workflow version.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.patch(
            "/".join([self.base_path, str(random_workflow.workflow_id), "versions", random_hex_string(), "status"]),
            headers=random_user.auth_headers,
            json=WorkflowVersionStatus(status=WorkflowVersion.Status.PUBLISHED).model_dump(),
        )

        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_deprecate_non_existing_workflow_version(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test deprecate a non-existing workflow version.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "versions", random_hex_string(), "deprecate"]),
            headers=random_user.auth_headers,
        )

        assert response.status_code == status.HTTP_404_NOT_FOUND


class TestWorkflowVersionRoutesGetDocumentation(_TestWorkflowVersionRoutes):
    @pytest.mark.asyncio
    async def test_download_workflow_version_documentation(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test downloading all the different documentation file for a workflow version.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        for document in DocumentationEnum:
            response = await client.get(
                "/".join(
                    [
                        self.base_path,
                        str(random_workflow.workflow_id),
                        "versions",
                        random_workflow.versions[0].git_commit_hash,
                        "documentation",
                    ]
                ),
                headers=random_user.auth_headers,
                params={"document": document.value},
            )
            assert response.status_code == status.HTTP_200_OK


class TestWorkflowVersionIconRoutes(_TestWorkflowVersionRoutes):
    @pytest.mark.asyncio
    async def test_upload_new_icon(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        mock_s3_service: MockS3ServiceResource,
        db: AsyncSession,
    ) -> None:
        """
        Test for uploading a new icon for a workflow version

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        """
        files = {"icon": ("RickRoll.txt", BytesIO(b"Never gonna give you up"), "plain/text")}
        version_hash = random_workflow.versions[0].git_commit_hash
        response = await client.post(
            "/".join(
                [
                    self.base_path,
                    str(random_workflow.workflow_id),
                    "versions",
                    version_hash,
                    "icon",
                ]
            ),
            headers=random_user.auth_headers,
            files=files,
        )
        assert response.status_code == status.HTTP_201_CREATED
        icon_slug = response.json()
        assert icon_slug in mock_s3_service.Bucket(settings.ICON_BUCKET).objects.all_keys()
        db_version = await db.scalar(select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == version_hash))
        assert db_version is not None
        assert db_version.icon_slug == icon_slug

        # Clean up
        mock_s3_service.Bucket(settings.ICON_BUCKET).Object(icon_slug).delete()

    @pytest.mark.asyncio
    async def test_delete_icon(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        random_workflow_version: WorkflowVersion,
        mock_s3_service: MockS3ServiceResource,
        db: AsyncSession,
    ) -> None:
        """
        Test for deleting a workflow version icon

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        random_workflow_version : clowmdb.models.WorkflowVersion
            Random workflow version for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        """
        old_slug = random_workflow_version.icon_slug
        response = await client.delete(
            "/".join(
                [
                    self.base_path,
                    str(random_workflow.workflow_id),
                    "versions",
                    random_workflow_version.git_commit_hash,
                    "icon",
                ]
            ),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT
        assert old_slug not in mock_s3_service.Bucket(settings.ICON_BUCKET).objects.all_keys()
        db_version = await db.scalar(
            select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == random_workflow_version.git_commit_hash)
        )
        assert db_version is not None
        assert db_version.icon_slug is None
