import asyncio
from functools import partial
from io import BytesIO
from secrets import token_urlsafe
from typing import AsyncIterator, Callable, Dict, Iterator

import httpx
import pytest
import pytest_asyncio
from clowmdb.db.session import get_async_session
from clowmdb.models import (
    Bucket,
    Workflow,
    WorkflowExecution,
    WorkflowMode,
    WorkflowVersion,
    workflow_mode_association_table,
)
from sqlalchemy import insert, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import get_db, get_decode_jwt_function, get_httpx_client, get_s3_resource
from app.core.config import settings
from app.git_repository import build_repository
from app.main import app
from app.schemas.workflow import WorkflowOut
from app.scm import SCM, Provider
from app.tests.mocks.mock_s3_resource import MockS3ServiceResource
from app.tests.utils.bucket import create_random_bucket
from app.tests.utils.user import UserWithAuthHeader, create_random_user, decode_mock_token, get_authorization_headers
from app.tests.utils.utils import handle_http_request, random_hex_string, random_lower_string

jwt_secret = token_urlsafe(32)


@pytest.fixture(scope="session")
def event_loop() -> Iterator:
    """
    Creates an instance of the default event loop for the test session.
    """
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def mock_s3_service() -> Iterator[MockS3ServiceResource]:
    """
    Fixture for creating a mock object for the rgwadmin package.
    """
    mock_s3 = MockS3ServiceResource()
    mock_s3.Bucket(settings.ICON_BUCKET).create()
    mock_s3.Bucket(settings.PARAMS_BUCKET).create()
    mock_s3.Bucket(settings.WORKFLOW_BUCKET).create()
    yield mock_s3
    mock_s3.delete_bucket(name=settings.ICON_BUCKET, force_delete=True)
    mock_s3.delete_bucket(name=settings.PARAMS_BUCKET, force_delete=True)
    mock_s3.delete_bucket(name=settings.WORKFLOW_BUCKET, force_delete=True)


@pytest_asyncio.fixture(scope="module")
async def client(mock_s3_service: MockS3ServiceResource, db: AsyncSession) -> AsyncIterator[httpx.AsyncClient]:
    """
    Fixture for creating a TestClient and perform HTTP Request on it.
    Overrides the dependency for the RGW admin operations.
    """

    def get_mock_s3() -> MockS3ServiceResource:
        return mock_s3_service

    def get_decode_token_function() -> Callable[[str], Dict[str, str]]:
        # Override the decode_jwt function with mock function for tests and inject random shared secret
        return partial(decode_mock_token, secret=jwt_secret)

    async def get_mock_httpx_client(raise_error: bool = False) -> AsyncIterator[httpx.AsyncClient]:
        # raises an 404 error if the query parameter 'raise_error' is true
        def mock_request_handler(request: httpx.Request) -> httpx.Response:
            return handle_http_request(request, raise_error)

        async with httpx.AsyncClient(transport=httpx.MockTransport(mock_request_handler)) as http_client:
            yield http_client

    app.dependency_overrides[get_httpx_client] = get_mock_httpx_client
    app.dependency_overrides[get_s3_resource] = get_mock_s3
    app.dependency_overrides[get_decode_jwt_function] = get_decode_token_function
    app.dependency_overrides[get_db] = lambda: db
    async with httpx.AsyncClient(app=app, base_url="http://localhost") as ac:
        yield ac
    app.dependency_overrides = {}


@pytest_asyncio.fixture(scope="module")
async def db() -> AsyncIterator[AsyncSession]:
    """
    Fixture for creating a database session to connect to.
    """
    async with get_async_session(
        url=str(settings.SQLALCHEMY_DATABASE_ASYNC_URI), verbose=settings.SQLALCHEMY_VERBOSE_LOGGER
    ) as dbSession:
        yield dbSession


@pytest_asyncio.fixture(scope="function")
async def random_user(db: AsyncSession) -> AsyncIterator[UserWithAuthHeader]:
    """
    Create a random user and deletes him afterwards.
    """
    user = await create_random_user(db)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_second_user(db: AsyncSession) -> AsyncIterator[UserWithAuthHeader]:
    """
    Create a random second user and deletes him afterwards.
    """
    user = await create_random_user(db)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_third_user(db: AsyncSession) -> AsyncIterator[UserWithAuthHeader]:
    """
    Create a random third user and deletes him afterwards.
    """
    user = await create_random_user(db)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_bucket(
    db: AsyncSession, random_user: UserWithAuthHeader, mock_s3_service: MockS3ServiceResource
) -> AsyncIterator[Bucket]:
    """
    Create a random user and deletes him afterwards.
    """
    bucket = await create_random_bucket(db, random_user.user)
    mock_s3_service.Bucket(name=bucket.name).create()
    yield bucket
    mock_s3_service.delete_bucket(name=bucket.name, force_delete=True)
    await db.delete(bucket)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_workflow(
    db: AsyncSession, random_user: UserWithAuthHeader, mock_s3_service: MockS3ServiceResource
) -> AsyncIterator[WorkflowOut]:
    """
    Create a random workflow and deletes it afterwards.
    """
    workflow_db = Workflow(
        name=random_lower_string(10),
        repository_url="https://github.de/example-user/example",
        short_description=random_lower_string(65),
        developer_id=random_user.user.uid,
    )
    db.add(workflow_db)
    await db.commit()
    icon_slug = random_hex_string()
    workflow_version = WorkflowVersion(
        git_commit_hash=random_hex_string(40),
        version="v1.0.0",
        _workflow_id=workflow_db.workflow_id.bytes,
        icon_slug=icon_slug,
    )
    db.add(workflow_version)
    await db.commit()
    mock_s3_service.Bucket(settings.ICON_BUCKET).Object(icon_slug).upload_fileobj(BytesIO(b"{}"))
    mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(f"{workflow_version.git_commit_hash}.json").upload_fileobj(
        BytesIO(b"{}")
    )
    yield WorkflowOut.from_db_workflow(db_workflow=workflow_db, versions=[workflow_version], load_modes=False)
    mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(f"{workflow_version.git_commit_hash}.json").delete()
    mock_s3_service.Bucket(settings.ICON_BUCKET).Object(icon_slug).delete()
    await db.delete(workflow_db)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_private_workflow(
    db: AsyncSession,
    random_workflow: WorkflowOut,
    random_workflow_version: WorkflowVersion,
    mock_s3_service: MockS3ServiceResource,
) -> AsyncIterator[WorkflowOut]:
    """
    Transform the random workflow into a private workflow.
    """
    # Update credentials in Database
    token = random_lower_string(15)
    username = random_lower_string(15)
    await db.execute(
        update(Workflow)
        .where(Workflow._workflow_id == random_workflow.workflow_id.bytes)
        .values(credentials_token=token, credentials_username=username)
    )
    await db.commit()

    # Upload SCM file to PARAMS_BUCKET
    scm_provider = Provider.from_repo(
        build_repository(
            random_workflow.repository_url, random_workflow_version.git_commit_hash, token=token, username=username
        ),
        name=f"repo{random_workflow.workflow_id.hex}",
    )
    assert scm_provider is not None
    with BytesIO() as f:
        SCM([scm_provider]).serialize(f)
        f.seek(0)
        mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"{random_workflow.workflow_id.hex}.scm").upload_fileobj(
            f
        )
    yield random_workflow
    # Delete SCM in PARAMS_BUCKET after tests
    mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"{random_workflow.workflow_id.hex}.scm").delete()


@pytest_asyncio.fixture(scope="function")
async def random_workflow_version(db: AsyncSession, random_workflow: WorkflowOut) -> WorkflowVersion:
    """
    Create a random workflow version. Will be deleted, when the workflow is deleted.
    """
    stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == random_workflow.versions[0].git_commit_hash)
    return await db.scalar(stmt)


@pytest_asyncio.fixture(scope="function")
async def random_workflow_execution(
    db: AsyncSession,
    random_workflow_version: WorkflowVersion,
    random_user: UserWithAuthHeader,
    mock_s3_service: MockS3ServiceResource,
) -> AsyncIterator[WorkflowExecution]:
    """
    Create a random workflow execution. Will be deleted, when the user is deleted.
    """
    execution = WorkflowExecution(
        user_id=random_user.user.uid, workflow_version_id=random_workflow_version.git_commit_hash, slurm_job_id=1
    )
    db.add(execution)
    await db.commit()
    await db.refresh(execution)
    mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"params-{execution.execution_id.hex}.json").upload_fileobj(
        BytesIO(b"{}")
    )
    yield execution
    mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"params-{execution.execution_id.hex}.json").delete()


@pytest_asyncio.fixture(scope="function")
async def random_workflow_mode(
    db: AsyncSession,
    random_workflow_version: WorkflowVersion,
    mock_s3_service: MockS3ServiceResource,
) -> AsyncIterator[WorkflowMode]:
    """
    Create a random workflow execution. Will be deleted, when the user is deleted.
    """
    mode = WorkflowMode(
        name=random_lower_string(8), schema_path=random_lower_string(32), entrypoint=random_lower_string(16)
    )
    db.add(mode)
    await db.commit()
    await db.execute(
        insert(workflow_mode_association_table),
        [
            {
                "workflow_version_commit_hash": random_workflow_version.git_commit_hash,
                "workflow_mode_id": mode.mode_id.bytes,
            }
        ],
    )
    await db.commit()
    key = f"{random_workflow_version.git_commit_hash}-{mode.mode_id.hex}.json"
    parameter_schema = mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(key)
    parameter_schema.upload_fileobj(BytesIO(b"{}"))
    yield mode
    parameter_schema.delete()
    await db.delete(mode)
    await db.commit()
