from datetime import datetime, timedelta

import pytest
from clowmdb.models import Bucket, BucketPermission
from sqlalchemy.ext.asyncio import AsyncSession

from app.crud import CRUDBucket
from app.tests.utils.bucket import add_permission_for_bucket
from app.tests.utils.user import UserWithAuthHeader
from app.tests.utils.utils import random_lower_string


class TestBucketCRUDCheckAccess:
    @pytest.mark.asyncio
    async def test_check_bucket_access_with_active_permission1(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with an active bucket permissions.
        Only the 'from' timestamp is set.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.user.uid, from_=datetime.now() - timedelta(weeks=1)
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_active_permission2(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with an active bucket permissions.
        Only the 'to' timestamp is set.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.user.uid, to=datetime.now() + timedelta(weeks=1)
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_active_permission3(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with an active bucket permissions.
        The 'from' and 'to' timestamp are set.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db,
            random_bucket.name,
            random_second_user.user.uid,
            to=datetime.now() + timedelta(weeks=1),
            from_=datetime.now() - timedelta(weeks=1),
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_inactive_permission1(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with an inactive bucket permissions.
        Only the 'from' timestamp is set.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.user.uid, from_=datetime.now() + timedelta(weeks=1)
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert not check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_inactive_permission2(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with an inactive bucket permissions.
        Only the 'to' timestamp is set.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.user.uid, to=datetime.now() - timedelta(weeks=1)
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert not check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_inactive_permission3(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with an inactive bucket permissions.
        The 'from' and 'to' timestamp are set.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db,
            random_bucket.name,
            random_second_user.user.uid,
            to=datetime.now() - timedelta(weeks=1),
            from_=datetime.now() - timedelta(weeks=2),
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert not check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_file_prefix_permission1(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with a bucket permissions.
        File prefix set on bucket permission and valid key.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(db, random_bucket.name, random_second_user.user.uid, file_prefix="some/prefix/")
        check = await CRUDBucket.check_access(
            db, random_bucket.name, random_second_user.user.uid, key="some/prefix/test.jpg"
        )
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_file_prefix_permission2(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with a bucket permissions.
        File prefix set on bucket permission and valid key.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(db, random_bucket.name, random_second_user.user.uid, file_prefix="some/prefix/")
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid, key="some/prefix/")
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_file_prefix_permission3(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with a bucket permissions.
        File prefix set on bucket permission and invalid key.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(db, random_bucket.name, random_second_user.user.uid, file_prefix="some/prefix/")
        check = await CRUDBucket.check_access(
            db, random_bucket.name, random_second_user.user.uid, key="some/other/prefix.jpg"
        )
        assert not check

    @pytest.mark.asyncio
    async def test_check_bucket_access_without_file_prefix_permission1(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with a bucket permissions.
        No File prefix set on bucket permission and valid key.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db,
            random_bucket.name,
            random_second_user.user.uid,
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid, key="some/prefix/")
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_access_without_file_prefix_permission2(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with a bucket permissions.
        File prefix set on bucket permission and no key.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(db, random_bucket.name, random_second_user.user.uid, file_prefix="some/prefix/")
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert not check

    @pytest.mark.asyncio
    async def test_check_bucket_access_as_bucket_owner(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket as the bucket owner.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user who has no permissions for the bucket.
        """
        check = await CRUDBucket.check_access(db, random_bucket.name, random_user.user.uid)
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_READ_Permission(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with only READ permission.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.user.uid, permission=BucketPermission.Permission.READ
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert not check

    @pytest.mark.asyncio
    async def test_check_bucket_access_with_WRITE_Permission(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket with only WRITE permission.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.user.uid, permission=BucketPermission.Permission.WRITE
        )
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert not check

    @pytest.mark.asyncio
    async def test_check_bucket_access_without_Permission(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for checking access to a bucket without permission.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user who has no permissions for the bucket.
        """
        check = await CRUDBucket.check_access(db, random_bucket.name, random_second_user.user.uid)
        assert not check


class TestBucketCRUDCheckBuckeExist:
    @pytest.mark.asyncio
    async def test_check_bucket_exist(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
    ) -> None:
        """
        Test for checking if bucket exist with an existing bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_bucket : clowmdb.models.Bucket
            Random bucket for testing.
        """
        check = await CRUDBucket.check_bucket_exist(db, random_bucket.name)
        assert check

    @pytest.mark.asyncio
    async def test_check_bucket_not_exist(
        self,
        db: AsyncSession,
    ) -> None:
        """
        Test for checking if bucket exist without an existing bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        """
        check = await CRUDBucket.check_bucket_exist(db, random_lower_string())
        assert not check
