import pytest
from clowmdb.models import WorkflowVersion
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from app.crud import CRUDWorkflowVersion
from app.schemas.workflow import WorkflowOut
from app.tests.utils.utils import random_hex_string


class TestWorkflowVersionCRUDGet:
    @pytest.mark.asyncio
    async def test_get_specific_workflow_version(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for getting a workflow version by its id from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow = await CRUDWorkflowVersion.get(db, random_workflow.versions[0].git_commit_hash)
        assert workflow is not None
        assert workflow.workflow_id == random_workflow.workflow_id

    @pytest.mark.asyncio
    async def test_get_specific_workflow_version_with_populated_workflow(
        self, db: AsyncSession, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for getting a workflow version by its id with populated workflow from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow = await CRUDWorkflowVersion.get(
            db, random_workflow.versions[0].git_commit_hash, populate_workflow=True
        )
        assert workflow is not None
        assert workflow.workflow_id == random_workflow.workflow_id
        assert workflow.workflow.workflow_id == random_workflow.workflow_id

    @pytest.mark.asyncio
    async def test_get_latest_unpublished_workflow_version(
        self, db: AsyncSession, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for getting the latest workflow version from a workflow which is not published from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow_version = await CRUDWorkflowVersion.get_latest(db, random_workflow.workflow_id, published=False)
        assert workflow_version is not None
        assert workflow_version.workflow_id == random_workflow.workflow_id
        assert workflow_version.git_commit_hash == random_workflow.versions[0].git_commit_hash

    @pytest.mark.asyncio
    async def test_get_latest_published_workflow_version(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for getting the latest workflow version from a workflow which is published from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow_version = await CRUDWorkflowVersion.get_latest(db, random_workflow.workflow_id)
        assert workflow_version is None

    @pytest.mark.asyncio
    async def test_get_all_workflow_versions(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for getting all versions from a workflow from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow_versions = await CRUDWorkflowVersion.list(db, random_workflow.workflow_id)
        assert len(workflow_versions) == 1
        assert workflow_versions[0].git_commit_hash == random_workflow.versions[0].git_commit_hash

    @pytest.mark.asyncio
    async def test_get_all_workflow_version_with_specific_status(
        self, db: AsyncSession, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for getting all versions with a specific status from a workflow from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow_versions = await CRUDWorkflowVersion.list(
            db, random_workflow.workflow_id, version_status=[random_workflow.versions[0].status]
        )
        assert len(workflow_versions) == 1
        assert workflow_versions[0].git_commit_hash == random_workflow.versions[0].git_commit_hash


class TestWorkflowVersionCRUDCreate:
    @pytest.mark.asyncio
    async def test_create_workflow_version(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for creating a workflow version in CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow_version = await CRUDWorkflowVersion.create(
            db,
            git_commit_hash=random_hex_string(),
            version="v2.0.0",
            wid=random_workflow.workflow_id,
            previous_version=random_workflow.versions[-1].git_commit_hash,
        )
        assert workflow_version is not None

        stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == workflow_version.git_commit_hash)
        created_workflow_version = await db.scalar(stmt)
        assert created_workflow_version is not None
        assert created_workflow_version == workflow_version


class TestWorkflowVersionCRUDUpdate:
    @pytest.mark.asyncio
    async def test_update_workflow_version_status(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for creating a workflow version in CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        await CRUDWorkflowVersion.update_status(
            db, git_commit_hash=random_workflow.versions[0].git_commit_hash, status=WorkflowVersion.Status.PUBLISHED
        )

        stmt = select(WorkflowVersion).where(
            WorkflowVersion.git_commit_hash == random_workflow.versions[0].git_commit_hash
        )
        version = await db.scalar(stmt)
        assert version
        assert version.status == WorkflowVersion.Status.PUBLISHED
