import random
import string
from typing import Dict

import httpx
from fastapi import status

from app.core.config import settings
from app.tests.mocks.authorization_service import handle_request as auth_handle_request
from app.tests.mocks.slurm_cluster import handle_request as slurm_handle_request


def random_lower_string(length: int = 32) -> str:
    """
    Creates a random string with arbitrary length.

    Parameters
    ----------
    length : int, default 32
        Length for the random string.

    Returns
    -------
    string : str
        Random string.
    """
    return "".join(random.choices(string.ascii_lowercase, k=length))


def random_hex_string(length: int = 40) -> str:
    """
    Creates a random string with only hexadecimal digits

    Parameters
    ----------
    length : int, default 32
        Length for the random string.

    Returns
    -------
    string : str
        Random string.
    """
    return "".join(random.choices("0123456789abcdef", k=length))


def random_ipv4_string() -> str:
    """
    Creates a random IPv4 address.

    Returns
    -------
    string : str
        Random IPv4 address.
    """
    return ".".join(str(random.randint(0, 255)) for _ in range(4))


def handle_http_request(request: httpx.Request, raise_error: bool = False) -> httpx.Response:
    """
    Handler for a mock HTTP request. Forwards it to the appropriate handler based on the requested URL.

    Parameters
    ----------
    request : httpx.Request
        Mock request.
    raise_error : bool, default False
        Flag if to return an 404 status code response.

    Returns
    -------
    response : httpx.Response
        Generated mock request.
    """
    url = str(request.url)
    if url.startswith(str(settings.OPA_URI)):
        request_body: Dict[str, str] = eval(request.content.decode("utf-8"))["input"]
        return auth_handle_request(body=request_body)
    elif url.startswith(str(settings.SLURM_ENDPOINT)):
        return slurm_handle_request(request.method)
    elif raise_error:
        return httpx.Response(status_code=status.HTTP_404_NOT_FOUND, json={})
    return httpx.Response(
        status_code=status.HTTP_200_OK,
        json={
            # When checking if a file exists in a git repository, the GitHub API expects this in a response
            "download_url": "https://example.com"
        },
    )
